"""Apply a tokenizer to the provided text fields"""

from abc import abstractmethod
from multiprocessing import cpu_count
from typing import List, Optional, Union

import datasets
import polars as pl
import pyarrow as pa

from bardi.nlp_engineering.utils import tokenizers_lib
from bardi.nlp_engineering.utils.polars_utils import retain_inputs
from bardi.nlp_engineering.utils.validations import validate_pyarrow_table, validate_str_cols
from bardi.pipeline import DataWriteConfig, Step


class TokenizerEncoder(Step):
    """The tokenizer encoder uses a trained tokenizer
    to split text into tokens.

    Note
    ----

    Avoid the direct instantiation of the TokenizerEncoder class
    and instead instantiate one of the child classes depending
    on hardware configuration.

    Attributes
    ----------

    fields : Union[str, List[str]],
        the name of the column containing a list of tokens
        that will be mapped to integers using a vocab
    return_tensors: str
        type of tensors to return, 'np' for Numpy arrays,
        'pt' for PyTorch tensors or 'tf' for TensorFlow
    concat_fields : bool
        whether the text fields should be concatenate into a single
        text field, defaults to False
    retain_input_fields : Optional[bool]
        If True, will retain the original contents of the fields specified in
        `fields` under the new names of: `TokenizerEncoder__<field>`
    retain_concat_field : Optional[bool]
        If True, will retain the concatenation of the fields specified in
        `fields` under the name specified in `field_rename` or 'text' if not
        specified. If `concat_fields` is not True, this parameter will
        have no effect
    field_rename : Optional[str]
        optional ability to rename the supplied field with
        the field_rename value
    hf_cache_dir : Optional[str]
        local directory where the HF pretrained tokenizers are stored
    model_name : Optional[str]
        name of a tokenizer file or folder. Not required if TokenizerTrainer
        is a prior step - tokenizer will be passed in artifacts
    cores : Optional[int]
        number of CPU cores for multithreading the tokenizer
    tokenizer_params : Optional[dict]
        provide fine-grained customization for any valid HuggingFace Tokenizer parameter
        through a dictionary
    tokenizer_model : transformers.PreTrainedTokenizerBase
        Tokenizer object passed through artifacts from TokenizerTrainer or
        read from file specified in `model_name`
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        return_tensors: str = "np",
        concat_fields: bool = False,
        retain_input_fields: bool = False,
        retain_concat_field: bool = False,
        field_rename: Optional[str] = None,
        hf_cache_dir: Optional[str] = None,
        model_name: Optional[str] = None,
        cores: Optional[int] = None,
        tokenizer_params: Optional[dict] = None,
    ):
        """Constructor method"""
        self.fields = fields
        if isinstance(fields, str):
            self.fields = [fields]
        self.field_rename = "text"
        if field_rename:
            self.field_rename = field_rename
        self.retain_input_fields = retain_input_fields
        self.retain_concat_field = retain_concat_field
        self.model_name = model_name
        if hf_cache_dir:
            self.model_name = f"{hf_cache_dir}/{model_name}"
        self.concat_fields = concat_fields
        self.cores = cpu_count()
        if cores:
            self.cores = cores
        self.tokenizer_params = {}
        if tokenizer_params:
            self.tokenizer_params = tokenizer_params
        self.return_tensors = return_tensors
        self.tokenizer_model = None
        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }  # Default write config

    @abstractmethod
    def run(self):
        """Abstract method"""
        pass


class CPUTokenizerEncoder(TokenizerEncoder):
    """Implementation of the TokenizerEncoder for CPU computation

    Note
    ----
    This implementation of the TokenizerEncoder is specific for CPU computation.

    Attributes
    ----------

    fields : Union[str, List[str]],
        the name of the column containing a list of tokens
        that will be mapped to integers using a vocab
    return_tensors: str
        type of tensors to return, 'np' for Numpy arrays,
        'pt' for PyTorch tensors or 'tf' for TensorFlow
    concat_fields : bool
        whether the text fields should be concatenate into a single
        text field, defaults to False
    retain_input_fields : Optional[bool]
        If True, will retain the original contents of the fields specified in
        `fields` under the new names of: `CPUTokenizerEncoder_input__<field>`
    retain_concat_field : Optional[bool]
        If True, will retain the concatenation of the fields specified in
        `fields` under the name specified in `field_rename` or 'text' if not
        specified. If `concat_fields` is not True, this parameter will
        have no effect
    field_rename : Optional[str]
        optional ability to rename the supplied field with
        the field_rename value
    hf_cache_dir : Optional[str]
        local directory where the HF pretrained tokenizers are stored
    model_name : Optional[str]
        name of a tokenizer file or folder. Not required if TokenizerTrainer
        is a prior step - tokenizer will be passed in artifacts
    cores : Optional[int]
        number of CPU cores for multithreading the tokenizer
    tokenizer_params : Optional[TokenizerConfig]
        provide fine-grained settings for applying tokenizer
    tokenizer_model : transformers.PreTrainedTokenizerBase
        Tokenizer object passed through artifacts from TokenizerTrainer or
        read from file specified in `model_name`
    """

    def __init__(self, *args, **kwargs):
        """Constructor method"""
        super().__init__(*args, **kwargs)

    def run(self, data: pa.Table, artifacts: dict = None) -> pa.Table:
        """Run a tokenizer encoder based on provided configuration

        The tokenizer encoder relies on receiving a tokenizer to apply to
        the text. The tokenizer can be supplied in multiple ways:
          - referencing the model_name at object creation
          - contained in the pipeline artifacts dictionary passed
            to the run method referenced by the key, 'tokenizer_model'

        Parameters
        ----------

        data : PyArrow Table
            The data to be processed. The data must contain the
            column specified by field at object creation
        artifacts : dict
            A dictionary of pipeline artifacts which contains
            a tokenizer referenced by the key, 'tokenizer_model'

        Returns
        -------

        Tuple[PyArrow Table, dict]
            The first element holding a PyArrow Table of data
            processed with the tokenizer encoder. The second
            element of the tuple intended for artifacts is None.

        Raises
        ------

        AttributeError
            The tokenizer wasn't supplied either at object creation
            or to the run method
        TypeError
            The run method was not supplied a PyArrow Table
        """

        # Perform validations
        # columns are of type string or List[str] -> pretokenization
        # was not applied
        validate_pyarrow_table(data=data)
        validate_str_cols(fields=self.fields, data=data)

        # Check if a tokenizer model was passed through the artifacts dict
        if artifacts:
            if "tokenizer_model" in artifacts.keys():
                self.tokenizer_model = artifacts["tokenizer_model"]

        if self.model_name:
            self.tokenizer_model = tokenizers_lib.load_tokenizer(self.model_name)

        if not self.tokenizer_model:
            raise AttributeError(
                "A tokenizer needs to be provided to this step. Either provide "
                "the model name in the `model_name` parameter at TokenizerEncoder "
                "instantiation or train a tokenizer in this pipeline before this "
                "step with the bardi TokenizerTrainer class."
            )

        # Set tokenizers params
        self.tokenizer_model = tokenizers_lib.set_tokenizer_params(
            self.tokenizer_model, **self.tokenizer_params
        )

        # Retain input columns in original form if desired
        df = pl.from_arrow(data)

        # Concat the fields if desired
        if self.concat_fields and len(self.fields) > 1:
            # columns are of type string, some of them are null
            # concat their values into a single string
            df = (
                df.pipe(
                    retain_inputs, self.retain_input_fields, self.fields, self.__class__.__name__
                )
                .with_columns(
                    pl.concat_str(
                        [pl.col(field).fill_null("") for field in self.fields],
                        separator=" ",
                    ).alias("text")
                )
                .with_columns(
                    [pl.col("text").str.replace_all(pattern=r"\s\s+", value=" ").str.strip_chars()]
                )
                .drop([field for field in self.fields])
            )

            # concat field is called 'X', rename if
            # other name provided, defaults to 'X'
            df = df.rename({"text": self.field_rename})

        # if concat was not run and a single field was provided
        # the field can be renamed
        elif len(self.fields) == 1:
            df = df.rename({self.fields[0]: self.field_rename})

        # Apply tokenizer - we could support more params
        def tokenize(text):
            # add check for len here or filter dataframe first
            return self.tokenizer_model(
                text,
                padding=True,
                truncation=True,
                max_length=self.tokenizer_model.model_max_length,
                return_tensors=self.return_tensors,
            )

        data = df.to_arrow()
        hf_dataset = datasets.Dataset(arrow_table=data)

        # there is a single column to process
        if self.concat_fields or len(self.fields) == 1:
            hf_dataset = hf_dataset.map(
                lambda row: tokenize(row[self.field_rename]),
                batched=True,
                num_proc=self.cores,
            )
            if not self.retain_concat_field:
                hf_dataset = hf_dataset.remove_columns(self.field_rename)

        # there are multiple columns each tokenized separately
        # with the same tokenizer
        else:
            for field in self.fields:
                hf_dataset = (
                    hf_dataset.map(
                        lambda row: tokenize(row[field]), batched=True, num_proc=self.cores
                    )
                    .rename_column("input_ids", f"{field}_input_ids")
                    .rename_column("attention_mask", f"{field}_attention_mask")
                    .remove_columns(field)
                )
        data = hf_dataset.data.table

        return (data, None)

    def get_parameters(self):
        """Retrive the post-processor object configuration
        Does not return the mapping (vocab) as it can be large

        Returns
        -------

        dict
            a dictionary representation of the post-processor's attributes
        """
        params = vars(self).copy()
        params["tokenizer_model"] = str(params["tokenizer_model"])

        return params



"""Train a tokenizer for transformer-based models"""

import os
from abc import abstractmethod
from typing import List, Tuple, TypedDict, Union

import polars as pl
import pyarrow as pa
from tokenizers import Tokenizer

from bardi.nlp_engineering.utils import tokenizers_lib
from bardi.nlp_engineering.utils.tokenizers_lib import TrainableTokenizer
from bardi.nlp_engineering.utils.validations import validate_pyarrow_table, validate_str_cols
from bardi.pipeline import DataWriteConfig, Step


class TokenizerTrainerArtifactsWriteConfig(TypedDict):
    """Indicates the keys and data types expected in an
    artifacts write config dict for the embedding generator
    if overwriting the default configuration."""

    vocab_format: str
    vocab_format_args: dict


class TokenizerTrainer(Step):
    """The TokenizerTrainer class provides ability to train:

    1) A NEW TOKENIZER FORM AN OLD ONE
        Train a new tokenizer based on a provided tokenizer (from_old_flag).
        Provide a trained tokenizers associated with given architecture
        (BERT, LLAMA) and train a new tokenizer from scratch that
        is configurated to the provided architecture.

    2) TRAIN NEW ARCHITECTURE AGNOSTIC TOKENIZER
        Use one of the supported tokenizer algorithms to train a new
        tokenizer from scratch.

    Note
    ----

    Avoid the direct instantiation of the TokenizerTrainer class
    and instead instantiate one of the child classes depending
    on hardware configuration.

    Attributes
    ----------

    field : str
        the name of the column containing text
    tokenizer_type: str
        types of tokenizers that can be trained from scratch
        currently supported WordPiece, BPE, Unigram, WordLevel
    vocab_size: int
        number of tokens in a trained tokenizer
    hf_cache_dir: str
        path to a folder where hf tokenizers are stored
    from_old_flag : bool
        if True, use pre-trained tokenizer as a template.
    checkpoint_path : str
        path to pretrained tokenizer model.
    tokenizer_fname: str
        name for the file or folder where the trained
        tokenizer will be stored
    corpus_gen_batch_size: int
        size of batch for tokenizer training data corpus
        by deafult it is 1000
    """

    def __init__(
        self,
        fields: Union[str, List[str]],
        tokenizer_type: str = "",
        vocab_size: int = 1000,
        hf_cache_dir: str = "",
        from_old_flag: bool = False,
        checkpoint_path: str = None,
        tokenizer_fname: str = "tokenizer",
        corpus_gen_batch_size: int = 1000,
        special_tokens: List[str] = None,
    ):
        """Constructor method"""
        if isinstance(fields, str):
            self.fields = [fields]
        else:
            self.fields = fields

        self.tokenizer_type = tokenizer_type
        self.vocab_size = vocab_size
        self.from_old_flag = from_old_flag
        self.checkpoint_path = checkpoint_path
        self.tokenizer_fname = tokenizer_fname
        self.corpus_gen_batch_size = corpus_gen_batch_size
        self.special_tokens = special_tokens

        if hf_cache_dir:
            self.checkpoint_path = f"{hf_cache_dir}/{checkpoint_path}"

        self._data_write_config: DataWriteConfig = {
            "data_format": "parquet",
            "data_format_args": {"compression": "snappy", "use_dictionary": False},
        }
        self._artifacts_write_config: TokenizerTrainerArtifactsWriteConfig = {
            "vocab_format": "json",
            "vocab_format_args": {},
        }

    def set_write_config(
        self,
        data_config: DataWriteConfig = None,
        artifacts_config: TokenizerTrainerArtifactsWriteConfig = None,
    ):
        """Overwrite the default file writing configurations"""

        if data_config:
            self._data_write_config = data_config
        if artifacts_config:
            self._artifacts_write_config = artifacts_config

    @abstractmethod
    def run(self):
        """Abstract method"""
        pass


class CPUTokenizerTrainer(TokenizerTrainer):
    """TransformerTokenizer specific for CPU computation.

    Note
    ----
    This implementation of the TokenizerTrainer is specific for CPU computation.

    Attributes
    ----------

    field : str
        the name of the column containing text
    tokenizer_type: str
        types of tokenizers that can be trained from scratch
        currently supported WordPiece, BPE, Unigram, WordLevel
    vocab_size: int
        number of tokens in a trained tokenizer
    hf_cache_dir: str
        path to a folder where hf tokenizers are stored
    from_old_flag : bool
        if True, use pre-trained tokenizer as a template.
    checkpoint_path : str
        path to pretrained tokenizer model.
    tokenizer_fname: str
        name for the file or folder where the trained
        tokenizer will be stored
    corpus_gen_batch_size: int
        size of batch for tokenizer training data corpus
        by deafult it is 1000
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.tokenizer_model = None

    def run(self, data: pa.Table, artifacts: dict) -> Tuple[pa.Table, dict]:
        """Runs tokenizer trainer based on the configuration used
        to create the object of the CPUTransformerTokenizer class

        Parameters
        ----------

        data : pyarrow.Table
            a pyarrow Table containing at least one list column containing
            text
        artifacts : dict
            artifacts are not consumed in this run method, but must be
            received to operate correctly in the pipeline run method

        Returns
        -------

        Tuple[pyarrow.Table, dict]
            The first position is a pyarrow.Table of pre-tokenized data.
            The second position is a dictionary of artifacts. The dict will
            contain keys for "embedding_matrix" and "id_to_token".
        """
        # Perform validations
        validate_pyarrow_table(data=data)
        validate_str_cols(fields=self.fields, data=data)

        # Use the Polars library to convert a text column into a list of lists
        # [each list here is a set of tokens]. The list of lists is called
        # df_words_lists. If there are multiple text columns, the df_words_list
        # is extended with contents of rows of the additional columns
        text_data = pl.from_arrow(data.select(self.fields))
        df_words_lists = []
        for field in self.fields:
            df_words_lists.extend(text_data[field].drop_nulls().to_list())

        # generator to iterate over all of the data
        # corpus geerator batch size
        def get_training_corpus():
            return (
                df_words_lists[i: i + self.corpus_gen_batch_size]
                for i in range(0, len(df_words_lists), self.corpus_gen_batch_size)
            )

        training_corpus = get_training_corpus()

        if self.from_old_flag:
            # load an existing tokenizer train a new one from the old one
            old_tokenizer = tokenizers_lib.load_hf_tokenizer(self.checkpoint_path)
            self.tokenizer_model = old_tokenizer.train_new_from_iterator(
                training_corpus, vocab_size=self.vocab_size
            )

        else:
            # train custom tokenizer from scratch, create a TrainableTokenizer
            # object with default values for special tokens and so on
            self.tokenizer_model = TrainableTokenizer(
                self.tokenizer_type, self.vocab_size, self.special_tokens
            )

            self.tokenizer_model.tokenizer.train_from_iterator(
                training_corpus, trainer=self.tokenizer_model.trainer
            )
            self.tokenizer_model = self.tokenizer_model.tokenizer

        # Set up the artifacts dict to return
        produced_artifacts = {
            "tokenizer_model": self.tokenizer_model,
            "vocab_size": self.vocab_size,
        }

        if self.from_old_flag:
            produced_artifacts["from_old_flag"] = self.checkpoint_path
        else:
            produced_artifacts["tokenizer_type"] = self.tokenizer_type

        return (data, produced_artifacts)

    def write_artifacts(self, write_path: str, artifacts: Union[dict, None]) -> None:
        """Write the oartifactsproduced by the embedding_generator

        Parameters
        ----------

        write_path : str
            Path is a directory where files will be written
        artifacts : Union[dict, None]
            Artifacts is a dictionary of artifacts produced in this step.
            Expected keys are: "id_to_token" and "embedding_matrix"
        """
        tokenizer_model = artifacts["tokenizer_model"]
        # Add the filename with the appropriate filetype to the write path

        if isinstance(tokenizer_model, Tokenizer):
            tokenizer_model_path = os.path.join(
                write_path,
                (f'{self.tokenizer_fname}.{self._artifacts_write_config["vocab_format"]}'),
            )
            tokenizer_model.save(tokenizer_model_path)

        else:
            tokenizer_model_path = os.path.join(write_path, f"{self.tokenizer_fname}")
            tokenizer_model.save_pretrained(tokenizer_model_path)

    def get_parameters(self) -> dict:
        """Retrive the embedding generetor object configuration

        Returns
        -------

        dict
            a dictionary representation of the embedding generetor object's attributes
        """

        params = vars(self).copy()
        params.pop("tokenizer_model")

        return params



from bardi.nlp_engineering.regex_library.regex_set import RegexSet, RegexSubPair
from bardi.nlp_engineering.regex_library.pathology_report import PathologyReportRegexSet
from bardi.nlp_engineering.regex_library.regex_lib import *



from bardi.nlp_engineering.embedding_generator import CPUEmbeddingGenerator
from bardi.nlp_engineering.label_processor import CPULabelProcessor
from bardi.nlp_engineering.normalizer import CPUNormalizer
from bardi.nlp_engineering.pre_tokenizer import CPUPreTokenizer
from bardi.nlp_engineering.regex_library import *
from bardi.nlp_engineering.splitter import CPUSplitter, MapSplit, NewSplit
from bardi.nlp_engineering.tokenizer_encoder import CPUTokenizerEncoder
from bardi.nlp_engineering.tokenizer_trainer import CPUTokenizerTrainer
from bardi.nlp_engineering.utils import validations
from bardi.nlp_engineering.utils.helper_utils import existing_split_mapping
from bardi.nlp_engineering.utils.tokenizers_lib import (TrainableTokenizer,
                                                        load_hf_tokenizer,
                                                        load_tokenizer,
                                                        set_tokenizer_params)
from bardi.nlp_engineering.vocab_encoder import CPUVocabEncoder



import json
from typing import List

import polars as pl


def existing_split_mapping(
    data_path: str, format: str, mapping_write_path: str, unique_record_cols: List[str]
):
    """Extract the split from an existing data_fold0.csv to
    duplicate the same split for pre-processing comparison

    Combines and hashes the columns in unique_record_cols,
    referred to as composite_record_id, and creates a json
    mapping {composite_record_id: split}

    Parameters
    ----------

    data_path : str
        full filepath to the existing data file
    format : str
        format of the existing file (currently only 'csv')
    mapping_write_path : str
        where the created mapping should be saved to
    unique_record_cols : List[str]
        the set of columns that create a distinct record
        if only one column, still provide in a list

    Returns
    -------
    dict
        dictionary assigning each unique row to a given split
    """
    if format == "csv":
        df = pl.scan_csv(source=data_path)
    elif format == "parquet":
        df = pl.scan_parquet(source=data_path)
    else:
        raise NotImplementedError("Support for this format has not been implemented yet.")

    data_fold0_df = (
        df.select(*unique_record_cols, "split")
        .with_columns(
            [
                pl.concat_str([*unique_record_cols])
                .hash()
                .cast(pl.Utf8())
                .alias("composite_record_id")
            ]
        )
        .select("composite_record_id", "split")
    ).collect()

    mappings = data_fold0_df.to_struct(name="mappings").to_list()

    mapping = {d["composite_record_id"]: d["split"] for d in mappings}

    with open(mapping_write_path, "w") as f:
        json.dump(mapping, f, indent=4)
    f.close()
    return mapping



"""Utilities for the tokenizers' support"""
from os.path import isdir
from typing import Union

from tokenizers import Tokenizer, models, pre_tokenizers, trainers
from transformers import AutoTokenizer, PreTrainedTokenizerFast, PreTrainedTokenizerBase


class TrainableTokenizer:
    """
    A class to reporesent a trainable tokenizer.
    It currently supports training the following
    HF tokenizers from scratch
    of type:
        - WordPiece
            subword tokenizer algorithm, uses greedy algorithm that tries
            to build long words first, splitting in multiple tokens when
            entire words do not exist in the vocabulary. Uses ## prefix
            to identify tokens that are part of a word (not starting a word)
        - BPE
            subword tokenization algorithm. BPE starts with characters
            then merges those that are seen next to each other most often,
            thus creating new tokens from most frequent pairs.
        - Unigram
            subword tokenization algorithm, works by trying to identify
            the best set of subword tokens to maximize the probability
            of a given sentence.
        - WordLevel
            simply maps words to IDs, required large vocabulary size
    """

    def __init__(self, tokenizer_type, voc_size, special_tokens):
        self.tokenizer_type = tokenizer_type
        self.voc_size = voc_size
        self.special_tokens = special_tokens
        self.tokenizer = None
        self.trainer = None
        self.subword_tokenizers = ["WordPiece", "BPE", "Unigram", "WordLevel"]

        if self.tokenizer_type == "WordPiece":
            # default special tokens for WordPiece tokenizer
            if self.special_tokens is None:
                self.special_tokens = ["[UNK]", "[PAD]", "[CLS]", "[SEP]", "[MASK]"]
            self.tokenizer = Tokenizer(models.WordPiece(unk_token="[UNK]"))

            self.trainer = trainers.WordPieceTrainer(
                vocab_size=self.voc_size,
                special_tokens=self.special_tokens,
            )

        elif self.tokenizer_type == "BPE":
            self.special_tokens = ["<|endoftext|>"]

            self.tokenizer = Tokenizer(models.BPE())
            self.trainer = trainers.BpeTrainer(
                vocab_size=self.voc_size,
                special_tokens=self.special_tokens,
            )

        elif self.tokenizer_type == "Unigram":
            self.special_tokens = ["<|endoftext|>"]

            self.tokenizer = Tokenizer(models.Unigram())
            self.trainer = trainers.UnigramTrainer(
                vocab_size=self.voc_size,
                special_tokens=self.special_tokens,
            )

        elif self.tokenizer_type == "WordLevel":
            self.special_tokens = ["<|endoftext|>"]

            self.tokenizer = Tokenizer(models.WordLevel())
            self.trainer = trainers.WordLevelTrainer(
                vocab_size=self.voc_size,
                special_tokens=self.special_tokens,
            )
        # subword tokenizers require a pretokenizer
        # for now all of them will use custom Whitespace
        if self.tokenizer_type in self.subword_tokenizers:
            self.tokenizer.pre_tokenizer = pre_tokenizers.Whitespace()

        return


def load_hf_tokenizer(path: str) -> AutoTokenizer:
    """
    Loads HF tokenizer for 'train a new tokenizer from an old' one option.

    Parameters
    ----------

    path : str
        path to a directory with HF tokenizer's config
        json files. We expect:
            - tokenizer.json
            - tokenizer_config.json
            - special_tokens_map.json

    Returns
    -------

    tokenizer : AutoTokenizer
        HF AutoTokenizer object
    """
    tokenizer = AutoTokenizer.from_pretrained(path)

    return tokenizer


def load_tokenizer(path: str) -> Union[Tokenizer, AutoTokenizer]:
    """
    Loads any HF tokenizer including model agnostic
    custom tokenizer. This will be extended to support
    SentencePiece tokenizers

    Parameters
    ----------

    path : str
        path to directory or to .json file
        with a tokenizer info

    Returns
    -------

    tokenizer : Union[Tokenizer, AutoTokenizer]
        return object of class Tokenizer
        or AutoTokenizer depending on the provided
        files
    """
    if isdir(path):
        return AutoTokenizer.from_pretrained(path)

    return Tokenizer.from_file(path)


def set_tokenizer_params(
    tokenizer: Union[AutoTokenizer, Tokenizer],
    model_max_length: int = 4096,
    unk_token: str = "[UNK]",
    pad_token: str = "[PAD]",
    *args, **kwargs
) -> AutoTokenizer:
    """Overwrites the defaults AutoTokenizer settings.
    If a user provides a TokenizerConfigs object then
    AutoTokenizer's settings are set to the values provided.
    If the user does not provide any values, the setting that
    come from tokenizer_config.json are used. If required values
    like model_max_length or pad_token are missing in the tokenizer_config.json
    then they are set to defaults.

    model_max_length = 4096
    unk_token = "[UNK]"
    pad_token = "[PAD]"

    Parameters
    ----------

    tokenizer : AutoTokenizer
        A HuggingFace tokenizer object
    model_max_length : int
        Defaults to 4096
    unk_token : str
        Defaults to
    pad_token : str
        Defaults to
    *args
        Variable length argument list
    **kwargs
        Arbitraty keyword argument list for any valid HuggingFace PreTrainedTokenizer
        parameters

    Returns
    -------

    tokenizer : AutoTokenizer
    """

    # Setting params for a custom trained tokenizer
    if isinstance(tokenizer, Tokenizer):
        tokenizer = PreTrainedTokenizerFast(
            tokenizer_object=tokenizer,
            max_length=model_max_length,
            unk_token=unk_token,
            pad_token=pad_token,
            *args, **kwargs
        )

    # Setting params for an HF PreTrainedTokenizer
    if isinstance(tokenizer, PreTrainedTokenizerBase):
        tokenizer.model_max_length = model_max_length

        # some pretrained tokenizers have model_max_size set
        # to  1000000000000000019884624838656 which is equivalent to not set
        if tokenizer.model_max_length > 1000000000000000:
            tokenizer.model_max_length = model_max_length

    # Add special tokens if they were not specified already
    if tokenizer.pad_token is None:
        tokenizer.add_special_tokens({"pad_token": pad_token})
    if tokenizer.unk_token is None:
        tokenizer.add_special_tokens({"unk_token": unk_token})

    return tokenizer



from typing import List

import polars as pl


def retain_inputs(
    df: pl.DataFrame, retain_input_fields: bool, fields: List[str], step_name: str
) -> pl.DataFrame:
    """Reusable function to apply each regex substitution
    normalization to each string field

    Parameters
    ----------

    df : pl.DataFrame
        a DataFrame containing the columns specified in the `fields` attribute
    retain_input_fields : bool
        should the original input fields be retained in a separate, renamed column
    fields : List[str]
        The name of the column(s) to be retained
    step_name : str
        The name of the step calling this function. Used to name the retained field.

    Returns
    -------
    pl.DataFrame
        a DataFrame with the specified fields
    """
    if retain_input_fields:
        df = df.with_columns(
            [pl.col(field).alias(f"{step_name}_input__{field}") for field in fields]
        )

    return df



from bardi.data.utils.pyarrow_utils import chunk_pyarrow_table
from bardi.data.data_handlers import Dataset
from bardi.data.data_handlers import (
    from_file,
    from_duckdb,
    from_pandas,
    from_pyarrow,
    from_json,
    to_pandas,
    to_polars,
    write_file
)



"""Define RegexSet and RegexSubPair blueprints"""

from typing import List, TypedDict


class RegexSubPair(TypedDict):
    """Dictionary used for regular expression string substitutions

    Example of a regex sub pair dictionary: ::

        {
            "regex_str": r"\\s",
            "sub_str": "WHITESPACE"
        }

    Attributes
    ----------

    regex_str : str
        regular expression pattern
    sub_str : str
        replacement value for matched string
    """

    regex_str: str
    sub_str: str


class RegexSet:
    """Blueprint for creating a configurable, domain specific regular expression set

    Attributes
    ----------

    regex_set : List[RegexSubPair]
        a list of regular expression substitution pairs
    """

    def __init__(self):
        self.regex_set: List[RegexSubPair] = []

    def get_regex_set(
        self, lowercase_substitution=False, no_substitution=False
    ) -> List[RegexSubPair]:
        """Return the ordered set of regular expressions

        Attributes
        ----------

        lowercase_substitution : Optional[bool]
            It True all the substitution tokeen like DATETOKEN will be returned
            in lowercase `datetoken`. Defaults to False.
        no_substitution : Optionl[bool]
            If True all the regular expression that remove matched pattern will replace it
            with space instead of special token. Defaults to False.

        Returns
        -------

        List[RegexSubPair]
            a list of regular expression substitution pairs
        """
        if no_substitution:
            for regex_pair in self.regex_set:
                if "TOKEN" in regex_pair["sub_str"]:
                    regex_pair["sub_str"] = " "

        elif lowercase_substitution:
            for regex_pair in self.regex_set:
                regex_pair["sub_str"] = regex_pair["sub_str"].lower()

        return self.regex_set
